// (C) 2017-2019 by folkert van heusden, released under AGPL v3.0
#include "config.h"
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "target_pixelflood.h"
#include "error.h"
#include "exec.h"
#include "log.h"
#include "picio.h"
#include "utils.h"

target_pixelflood::target_pixelflood(const std::string & id, const std::string & descr, source *const s, const double interval, const std::vector<filter *> *const filters, const double override_fps, configuration_t *const cfg, const std::string & host, const int port, resize *const r, const int pfw, const int pfh, const int quality, const pixelflood_protocol_t pp, const int xoff, const int yoff, const bool use_previous_when_fail) : target(id, descr, s, "", "", fmt, max_time, interval, filters, "", "", "", override_fps, cfg, false, use_previous_when_fail), quality(quality), host(host), port(port), r(r), pw(pfw), ph(pfh), pp(pp), xoff(xoff), yoff(yoff)
{
	if (this -> descr == "")
		this -> descr = store_path + "/" + prefix;

	if (pp == PP_TCP_TXT) {
		fd = connect_to(host, port, nullptr);
	}
	else if (pp == PP_UDP_BIN) {
		fd = socket(AF_INET, SOCK_DGRAM, 0);

		memset(&servaddr, 0, sizeof(servaddr));
		servaddr.sin_family = AF_INET;
		servaddr.sin_port = htons(port);
		servaddr.sin_addr.s_addr = inet_addr(host.c_str());
	}
	else
		log(LL_ERR, "No protocol selected");

	if (fd == -1)
		log(LL_ERR, "Cannot create socket");
}

target_pixelflood::~target_pixelflood()
{
	stop();

	close(fd);
}

bool target_pixelflood::send_frame(const uint8_t *const data, const int w, const int h)
{
	bool rc = true;
	char buffer[65507];
	int o = 0;

	if (pp == PP_UDP_BIN) {
		buffer[0] = 0; // protocol version 0
		buffer[1] = 0;
		o = 2;
	}

	for(int y=0; y<ph - yoff; y++) {
		for(int x=0; x<pw - xoff; x++) {
			const uint8_t *const p = &data[y * w * 3 + x * 3];

			int cx = x + xoff;
			int cy = y + yoff;

			if (pp == PP_UDP_BIN) {
				buffer[o++] = cx;
				buffer[o++] = cx >> 8;
				buffer[o++] = cy;
				buffer[o++] = cy >> 8;
				buffer[o++] = p[0];
				buffer[o++] = p[1];
				buffer[o++] = p[2];

				if (o >= 1122 - 6) {
					if (sendto(fd, buffer, o, 0, (const struct sockaddr *) &servaddr, sizeof(servaddr)) == -1)
						log(LL_INFO, "Cannot send pixelflood packet (%s)", strerror(errno));

					o = 2;
				}
			}
			else if (pp == PP_TCP_TXT) {
				o += sprintf(&buffer[o], "PX %d %d %02x%02x%02x\n", cx, cy, p[0], p[1], p[2]);

				if (o >= sizeof(buffer) - 100) {
					if (WRITE(fd, buffer, o) != o) {
						log(LL_INFO, "Cannot send pixelflood packet (%s)", strerror(errno));
						rc = false;
					}

					o = 0;
				}
			}
		}
	}

	if (pp == PP_UDP_BIN) {
		if (o > 2 && sendto(fd, buffer, o, 0, (const struct sockaddr *) &servaddr, sizeof(servaddr)) == -1)
			log(LL_INFO, "Cannot send pixelflood packet (%s)", strerror(errno));
	}
	else if (pp == PP_TCP_TXT) {
		if (o && WRITE(fd, buffer, o) != o) {
			log(LL_INFO, "Cannot send pixelflood packet (%s)", strerror(errno));
			rc = false;
		}
	}

	return rc;
}

void target_pixelflood::operator()()
{
	set_thread_name("pixelflood_" + prefix);

	s -> register_user();

	time_t cut_ts = time(nullptr) + max_time;

	uint64_t prev_ts = 0;
	bool is_start = true;
	std::string name;
	unsigned f_nr = 0;

	uint8_t *prev_frame = nullptr;

	for(;!local_stop_flag;) {
		pauseCheck();

		double before_ts = get_ts();

		int w = -1, h = -1;
		uint8_t *work = nullptr;
		size_t work_len = 0;
		if (s -> get_frame(E_RGB, quality, &prev_ts, &w, &h, &work, &work_len, use_previous_when_fail)) {
			source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
			instance_t *inst = find_instance_by_interface(cfg, cur_s);

			apply_filters(inst, cur_s, filters, prev_frame, work, prev_ts, w, h);

			bool rc = true;

			if (w != pw || h != ph) {
				uint8_t *temp = nullptr;
				r -> do_resize(w, h, work, pw, ph, &temp);

				rc = send_frame(temp, pw, ph);

				free(temp);
			}
			else {
				rc = send_frame(work, w, h);
			}

			if (!rc) {
				close(fd);

				fd = connect_to(host, port, nullptr);
			}

			free(prev_frame);
			prev_frame = work;
		}

		handle_fps(before_ts);
	}

	free(prev_frame);

	s -> unregister_user();

	log(id, LL_INFO, "stopping");
}
